<?php
/*
* Box Plot (Box and Whisker Graph) PHP5 class v1.0 
* Copyright (c) 2007 ismail ata kurt
* http://blog.livaxmedia.com
*/

class boxplot {
	
	const red		=0xF01E3C;
	const green 	=0x38D878;
	const blue		=0x1717F6;	    
    const orange	=0xF6AF17;
    const yellow	=0xEBFF06;
    const pink		=0xEC80DE;
    const turquaz	=0x06FFFC;
    const grey		=0xCCCCCC;
    const black		=0x000000;  
    const white		=0xFFFFFF; 
    
	private $output;
    private $X;
   	private $Y;				
   	private $font;
   	private $font_width;
   	private $vertical_values;
   	private $horizontal_values;
   	
    private $margin_top		 	= 20;
    private $margin_left;
    private $box_width;
    private $box_margin_left	= 25;    
    private $debug				=false;
    
    
    function __construct($height=200, 			  //Default height of the graph
    						$box_width=18	,		
    						$font="VeraMono.ttf" ,//nice fixed size font (http://www.dafont.com/bitstream-vera-mono.font) 
    						$font_witdh=8,
    						$output=NULL,
    						$debug=false
    						) {    	
    	
    	
    	if(isset($height)) $this->Y=$height;    	
    	if(isset($box_width)) $this->box_width=$box_width;
    	if(isset($font)) $this->font=$font;    	
    	if(isset($font_witdh)) $this->font_width=$font_witdh;
    	if(isset($output)) $this->output=$output;
    	if($debug) $this->debug=true;   	
    	
    }
    
    /*
    * Self testing for class
    */
    public function self_test() {     	
    	
    	$data = array(  'A' => array(11,11.2,11.5,11.6,11.9,12,12.2,12.8,12.9,13.1) ,
            	        'B' => array(13.3,13.4,13.8,13.9,14.2,14.5,14.5,14.6,15.3,15.5,16,19) ,
                	    'C' => array(8,12,13,14,15.1,15.4,15.4,15.7,16,16.7,17,17.1,17.4,17.9,18,20) ,
                	    'D' => array(8,8.1,8.8,9,9.7,9.8,9.9,10,12.1,13,14,19),
                	    'E' => array(9,9.1,9.2,9.3,9.4,9.8,10,11,11.2,11.4,12,12.2,12.3,12.4,12.5,13,19)            	    
                    	);
    
    	$this->Draw('Values','Groups',$data);
	}
	
	
    /* parameters:
    *		$vV	  -> Vertical values text
    *		$hV	  -> Horizontal values text
    *		$data -> array of data sets
	*/
	public function Draw($vV,$hV,$data) {   
        
    	$Mins = array();
    	$Maxs = array();    	
    	$N=0;
    	
    	foreach($data as $dt) {
    		sort($dt,SORT_NUMERIC);
    		array_push($Mins,$dt[0]);
    		array_push($Maxs,$dt[sizeof($dt)-1]);
    		if(sizeof($dt)>$N) $N=sizeof($dt);
    	}
    	
    	$min = min($Mins);
    	$max = max($Maxs);
    	
    	$this->margin_left=(12+(strlen(intval($max))+1)*$this->font_width);
    	
    	
    	
    	$this->X=$this->margin_left+5 + sizeof($data)*($this->box_width+(abs($this->box_width-$this->box_margin_left))) + 25;
    	
    	$X=$this->X;
    	$Y=$this->Y;  	
    	
    	$img = ImageCreateTrueColor($X,$Y);
    	imagefill($img,0,0,self::white);    
    	imagerectangle($img,0,0,$X-1,$Y-1,self::turquaz); 
    
    	imagettftext($img, 8, 90, 12, ($Y-40), self::grey, $this->font, $vV);
    	imagettftext($img, 8, 0, 40, ($Y-10), self::grey, $this->font, $hV);
    
    	imageline($img,$this->margin_left,20,$this->margin_left,$Y-40,self::black); //vertical line
    	imageline($img,$this->margin_left,$Y-40,$X-20,$Y-40,self::black); //horizontal line    
    
    	$working_area_y = $Y-70;
    
    	
    	$key = array_keys($data);
    	$i=0;
    	foreach($data as $dt) {
        	unset($result);
        	$result = $this->calcStats($dt);         
        	
        	$key_name= $key[$i];        	
        	$i++;
        
        	$this->Draw_BoxAndWhisker($img ,
					array ( 
					    				'num'  => $i	,
                    	        		'box_width' => $this->box_width	,
                        	            	'wa'        => $working_area_y 
			 			),
					array( 
                    	                'min'		=> $min		  ,
                        	            'max'		=> $max		  ,
										'sno'       => $result[1] ,
                                	    'lno'       => $result[2] ,
                                    	'avg'       => $result[3] ,
                                    	'q1'        => $result[4] ,
                                    	'q2'        => $result[5] ,
                                    	'q3'        => $result[6] 
                            		)
        	);
        	
			
        	$outlierArray= $result[8];	
        	$lowerOutliers = $outlierArray[2];
        	$upperOutliers = $outlierArray[3];
        	
        	if(!empty($lowerOutliers)) $this->Draw_Outliers($img,$lowerOutliers,$min,$max,$result[4],$result[6],$result[7],$working_area_y,$i);
        	if(!empty($upperOutliers)) $this->Draw_Outliers($img,$upperOutliers,$min,$max,$result[4],$result[6],$result[7],$working_area_y,$i);
        	        	
        	$this->Draw_DatasetName($img,$key_name,$i,18);
        	
			if($this->debug) {
        		echo "<br/>";
        		echo "n          =".$result[0]."<br/>";
        		echo "sno        =".$result[1]."<br/>";
        		echo "lno        =".$result[2]."<br/>";
        		echo "avg        =".$result[3]."<br/>";
        		echo "q1         =".$result[4]."<br/>";
        		echo "median     =".$result[5]."<br/>";
        		echo "q3         =".$result[6]."<br/>";
        		echo "IQR        =".$result[7]."<br/>";
        		echo "<hr>";
			}
             
    	}     
		
    	$range_span = $this->range_span($max,$min,$N);
    	if($this->debug)  {
    		echo "<br/>min = $min   max = $max <br/>";
    		echo "<br/>SPAN = ".$range_span.'<br/>';
    	}  
	    
	    $this->Draw_LineValues($img,$min,$max,$range_span);
    
	    imagepng($img,$this->output);
    	imagedestroy($img);    	
   
    
	}

    /*
    * Draw each Box and Whisker
    */
    private function Draw_BoxAndWhisker(&$img,$params,$stats) {
           
    	   	
    	$margin_top = $this->margin_top;
    	$box_margin_left = $this->box_margin_left;
    	$offset_x = $this->margin_left + 5 + (($params['num']-1)*$box_margin_left);    
    	
    	//q3 y position
    	$q3=$margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['q3']);    	
    	//q1 y position
    	$q1=$margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['q1']);    
    	//q2 y position
    	$q2=$margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['q2']);    
    	//avg y position
    	$avg = $margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['avg']);    
    	//sno (CL)
    	$sno=$margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['sno']);    
    	//lno (CU)
    	$lno=$margin_top+$this->scale($stats['max'],$stats['min'],0,$params['wa'],$stats['lno']);   
    	//IQR
    	imagerectangle($img,$offset_x,$q3,$offset_x+$params['box_width'],$q1,black);    
    	//q2 (median)
    	imageline($img,$offset_x,$q2,$offset_x+$params['box_width'],$q2,black);   
    	
    	$lx =$offset_x+($params['box_width']/2);
    
    	//sno 
    	imageline($img,$lx-5,$sno,$lx+5,$sno,self::blue);
    	imagedashedline($img,$lx,$sno,$lx,$q1,self::black);
    
    	//lno   
    	imageline($img,$lx-5,$lno,$lx+5,$lno,self::red);
    	imagedashedline($img,$lx,$lno,$lx,$q3,self::black);
    
    	//average
    	imagefilledrectangle($img,$lx-1,$avg-1,$lx+1,$avg+1,self::green);    	
    	
    	if($this->debug) {
    		echo '<br/>';	
    		echo "q3 (scaled) = $q3<br/>";
    		echo "q1 (scaled) = $q1<br/>";
    		echo "q2 (scaled) = $q2<br/>";
    		echo "sno (scaled) = $sno<br/>";
    		echo "lno (scaled) = $lno<br/>";    		
    	}
    
	}
	
	private function Draw_DatasetName(&$img,$name,$num,$box_width) {		
    	$box_margin_left = $this->box_margin_left;    	
    	$offset_x = $this->margin_left + 5 + (($num-1)*$box_margin_left);    	
		$lx =$offset_x+($box_width/2);		
		imagettftext($img,8,0,$lx,$this->Y-25,self::black,$this->font,$name);		
	}
    
    private function Draw_LineValues(&$img,$min,$max,$span) {
		$s_min= $min;
		$s_max= $max;
		$max = intval($max);
		$min = intval($min);	
		
		(($min%$span)==0)?$start=$min:$start=$min+($min%$span);
		
		for($i=$start;$i<=$max;$i+=$span) {
			$y = $this->margin_top + $this->scale($s_max,$s_min,0,$this->Y-70,$i);						
			imageline($img,$this->margin_left,$y,$this->margin_left-2,$y,self::black);			
			imagettftext($img, 8, 0, 15, $y, self::black, $this->font, $i);
		}
	}
	
	/*
	* Draws mid,extreme outliers of the data set
	*/	
	private function Draw_Outliers(&$img,$data,$min,$max,$q1,$q3,$IQR,$wa,$num) {
		
		$box_margin_left = $this->box_margin_left;    	
    	$offset_x = $this->margin_left + 5 + (($num-1)*$box_margin_left);    	
		$lx =$offset_x+($this->box_width/2);		
				
		foreach($data as $dt) {
			
			if(($dt>=($q1-3*$IQR) && $dt <=($q1-1.5*$IQR)) || ($dt>=($q3+1.5*$IQR) && $dt <=($q3+3*$IQR))) {
				//mid outlier				
				$y =$this->margin_top+$this->scale($max,$min,0,$wa,$dt);
				imagettftext($img, 8, 0, $lx-3, $y, self::black , $this->font, "*");
				
			}			
			
			if($dt<($q1-3*$IQR) || $dt >($q3+3*$IQR)) {
				//ectreme outlier
				$y =$this->margin_top+$this->scale($max,$min,0,$wa,$dt);
				imagettftext($img, 8, 0, $lx-3, $y, self::black , $this->font, "o");
			}
			
		}
		
	}
		
    /* 
    * Calculates the required statistics for drawing box plots
    * 
    * parameters:
    *	data-> data array 
    *
    * returns array        
    *	=> size of data
    *	=> minimum
    *	=> maximum
    *	=> average
    *	=> q1
    *	=> q2
    *	=> q3
    *	=> IQR
    *	=> outliers array
	* 	     => smallest non-outlier observation
    *        => largest non-outlier observation
    *        => array of lower outliers
    *        => array of upper  outliers
	*/
	private	function calcStats(&$data) {
    
    	sort($data,SORT_NUMERIC);
    
    	$n          = sizeof($data);
    	$min        = min($data);
    	$max        = max($data);
    	$average    = array_sum($data)/$n;
    	$q1         = $this->quartile($data,1);
    	$q2         = $this->quartile($data,2);
    	$q3         = $this->quartile($data,3);
    	$IQR        = $q3-$q1;                                	  //interquartile range
    	$outliers   = $this->outliers($data,$q1,$q3,$IQR,$min,$max);
    	$sno        = $outliers[0];                             //smallest non-outlier observation
    	$lno        = $outliers[1];                            //largest non-outlier observation      
    	
    	return  array($n,$sno,$lno,$average,$q1,$q2,$q3,$IQR,$outliers);   
	}
	
	/*
	* Scales the calculated statistics for graphical presentation
	*/
	private	function scale($up1,$down1,$up2,$down2,$value) {    
	    return ((($value-$down1)*($up2-$down2))/($up1-$down1) ) + $down2;
	}

    /*
    * Calculate the range span for values
	*/
	private	function range_span($max,$min,$n) {
    	$temp = ($max-$min) / intval(10*log10($n));    
    	
    	if($this->debug) {
    		echo "temporary span value = $temp<br/>";
    	}   		
    	
    		for($i=-10;$i<=10;$i++) {    			
    			$rslt = pow(10,$i);  
    			
    			if($rslt==$temp) return $temp;  		
    			
    			if($temp < $rslt) {
    				if($temp >pow(10,($i-1))) {
    					return  pow(10,$i);
    				}
    			}
    		}    	
    		
	}
	    	
    /*
    * Finds Outlier values
    *
    * returns array 
    *        => smallest non-outlier observation,
    *        => largest non-outlier observation,
    *        => array of lower outliers ,
    *        => array of upper  outliers
	*/    
	private function outliers($data,$q1,$q3,$IQR,$min,$max) {
    	$lowerBound = $q1-1.5*$IQR;
    	$upperBound = $q3+1.5*$IQR;
    
    	if($this->debug) {
    		echo "FL=$lowerBound<br/>";
    		echo "FU=$upperBound<br/>";
    	}
    	
    	$lOutliers = array();   
    	$uOutliers = array();    	
    	
    	foreach($data as $value) {
        	if($value < $lowerBound) {
            	array_push($lOutliers,$value);
        	}
        	if($value > $upperBound) {
            	array_push($uOutliers,$value);
        	} 
    	}
    	(!empty($lOutliers))?$sno=$data[sizeof($lOutliers)]:$sno=$min;
    	(!empty($uOutliers))?$lno=$data[(sizeof($data)-sizeof($uOutliers))-1]:$lno=$max;
    
    	return array($sno,$lno,$lOutliers,$uOutliers);
	}
    
	/*
    * Calculate the Quartile
	*/
    private function quartile($data,$num) {
    	$data_size = sizeof($data);
    	switch($num) {
        	case 1:
            	//First quartile
            	return $data[round(0.25*($data_size+1))];
        	break;
        	case 2:
            	//Second quartile (median)
            	if($data_size%2==0){
                	 return ($data[floor($data_size/2)-1] + $data[floor($data_size/2)])/2;
            	}else {
                	 return $data[($data_size-1)/2];
            	}
        	break;
        	case 3:
            	//Third quartile
            	return $data[round(0.75*($data_size+1))];
        	break;        
        	default:
            	die("Bad quartile number");
    	}
	}

}